import logging
from pathlib import Path

import pandas as pd

from . import context_sources

__all__ = ["ServiceContext", "ProcessContext"]

        
class ServiceContext(context_sources.ServiceContextSource):
    def __init__(self, context_source: context_sources.ServiceContextSource, logger: logging.Logger) -> None:
        super().__init__()

        self._context_source = context_source
        self._logger = logger

    def get_parameter_value(self, name: str, required: bool = True, default: str = None) -> str:
        return self._context_source.get_parameter_value(name, required=required, default=default)

    def get_parameter_real_path_value(self, name: str, required: bool = True) -> Path:
        return self._context_source.get_parameter_real_path_value(name, required=required)
    
    @property
    def logger(self):
        return self._logger
    
class ProcessContext(context_sources.ProcessContextSource):
    def __init__(self, context_source: context_sources.ProcessContextSource, logger: logging.Logger) -> None:
        super().__init__()

        self._context_source = context_source
        self._logger = logger

    def get_parameter_value(self, name: str, required: bool = True, default: str = None) -> str:
        return self._context_source.get_parameter_value(name, required=required, default=default)
    
    def get_input_dataframe(self, name: str, required: bool = True) -> pd.DataFrame:
        return self._context_source.get_input_dataframe(name, required=required)

    def set_output_dataframe(self, name: str, df: pd.DataFrame):
        return self._context_source.set_output_dataframe(name, df)

    @property
    def logger(self):
        return self._logger
    


import typing

from mlservicewrapper import config


class ParameterSchema:
    def __init__(self, name: str, d: dict) -> None:
        self.name = name
        self.required = d.get("required", True)
        self.type = d["type"]

class ParametersSchema:
    def __init__(self, d: typing.Dict[str, typing.Any]) -> None:
        self._data = d

    def items(self) -> typing.Iterable[ParameterSchema]:
        return (ParameterSchema(k, v) for k, v in self._data.items())

class DatasetSchema:
    def __init__(self, name: str, d: dict) -> None:
        self.name = name
        self.required = d.get("required", True)
        self.dataset_type = d.get("datasetType")
        self.item_schema = d.get("itemSchema")

class ServiceSchema:
    def __init__(self, schema_spec: config.ConfigurationPart) -> None:
        self._schema_spec = schema_spec

    def _get_parameters_schema(self, name: str):
        p: typing.Dict[str, dict] = self._schema_spec.get_dict_value("parameters") or dict()

        ret = p.get(name, dict())

        return ParametersSchema(ret)

    def _get_dataset_specs(self, direction: str) -> typing.Iterable[DatasetSchema]:
        if self._schema_spec is None:
            return list()

        datasets: typing.Dict[str, typing.Dict[str, dict]] = self._schema_spec.get_dict_value("datasets") or dict()

        if datasets is None:
            return list()

        dataset_dict = datasets.get(direction)

        if dataset_dict is None:
            return list()

        return (DatasetSchema(k, v) for k, v in dataset_dict.items())
        
    def parameters_load(self):
        return self._get_parameters_schema("load")
        
    def parameters_process(self):
        return self._get_parameters_schema("process")

    def datasets_input(self):
        return self._get_dataset_specs("input")
        
    def datasets_output(self):
        return self._get_dataset_specs("output")

class ServiceInfo:
    def __init__(self, vals: config.ConfigurationPart) -> None:
        self.name = vals.get_value("name")
        self.version = vals.get_value("version")

class ServiceHostParameters:
    def __init__(self, vals: config.ConfigurationPart) -> None:
        self._vals = vals

    def get_host_config(self, name: str):
        return self._vals.get_sub_config(name, is_nested=True, required=False)

class ServiceParameters:
    def __init__(self, config_part: config.ConfigurationPart) -> None:
        self._config = config_part

    def get_value(self, name: str):
        return self._config.get_value(name)

    def get_real_path_value(self, name: str):
        return self._config.get_fully_qualified_path(name)

class ServiceLoadParameters:
    def __init__(self, config: config.ConfigurationPart) -> None:
        self.module_path = config.get_fully_qualified_path("modulePath")
        self.module_name = config.get_value("moduleName")
        self.package_name = config.get_value("packageName")

        if self.module_name and self.module_path:
            raise ValueError("Both modulePath and moduleName were specified!")
        elif not self.module_name and not self.module_path:
            raise ValueError("The module couldn't be determined! Please include either a modulePath or moduleName in configuration.")

        self.class_name = config.get_value("className")
        self.service_instance_name = config.get_value("serviceInstanceName")

        if self.class_name and self.service_instance_name:
            raise ValueError("Both className and serviceInstanceName were specified!")
        elif not self.class_name and not self.service_instance_name:
            raise ValueError("Either className or serviceInstanceName must be specified!")


T = typing.TypeVar("T")
class ServiceConfiguration:
    def __init__(self, config_file: str, host_type: str) -> None:
        self._config = config.from_file(config_file)

        if host_type:
            hosts = self._config.get_sub_config("host", is_nested=True, required=False)

            self._config = hosts.get_sub_config(host_type, is_nested=True, required=False)

    def _get_typed(self, type: typing.Callable[[config.ConfigurationPart], T], name: str, coalesce = False) -> T:
        d = self._config.get_sub_config(name)

        if d:
            return type(d)
        if coalesce:
            return type(config.empty())
        else:
            return None

    def plugins(self):
        parts = self._config.get_multi_sub_config("plugins")

        for part in parts:
            yield ServiceLoadParameters(part)

    def service(self):
        return ServiceLoadParameters(self._config)

    def schema(self):
        return self._get_typed(ServiceSchema, "schema", coalesce=True)

    def info(self):
        return self._get_typed(ServiceInfo, "info", coalesce=True)

    def host(self):
        return self._get_typed(ServiceHostParameters, "host", coalesce=True)

    def parameters(self):
        return ServiceParameters(self._config.get_sub_config("parameters"))



import inspect
import logging
import os
import typing
from pathlib import Path

from . import configuration, context_sources, contexts, errors, services
from .internal import service_loading as service_loading

__all__ = ["ServerInstance"]

class _ServiceConfigurationServiceContextSource(context_sources.ServiceContextSource):
    def __init__(self, config: configuration.ServiceConfiguration):
        self.__parameters = config.parameters()

    def get_parameter_value(self, name: str, required: bool = True, default: str = None) -> str:
        context_sources.NameValidator.raise_if_invalid(name)

        val = self.__parameters.get_value(name) or default
        
        if required and val is None:
            raise errors.MissingParameterError(name)

        return val

    def get_parameter_real_path_value(self, name: str, required: bool = True) -> Path:
        context_sources.NameValidator.raise_if_invalid(name)

        val = self.__parameters.get_real_path_value(name)
        
        if required and val is None:
            raise errors.MissingParameterError(name)

        return val

class SafeServiceWrapper(services.Service):
    def __init__(self, service: services.Service) -> None:
        self._service = service
        self._is_loaded = False

    async def _as_coroutine(self, val):
        
        if inspect.iscoroutine(val):
            return await val
        else:
            return val

    def has_load(self):
        return hasattr(self._service, "load")

    async def load(self, ctx: contexts.ServiceContext = None):
        load = getattr(self._service, 'load', None)

        if load:
           await self._as_coroutine(load(ctx))

        self._is_loaded = True

    async def process(self, ctx: contexts.ProcessContext):
        if not self._is_loaded and self.has_load():
            raise ValueError("Be sure to call load before process!")
        
        await self._as_coroutine(self._service.process(ctx))

    def dispose(self):
        disp = getattr(self._service, 'dispose', None)

        if disp:
            disp()
    
class ServerInstance:
    def __init__(self, host_type: str, config_path: str = None):

        if not config_path:
            config_path = os.environ.get("SERVICE_CONFIG_PATH", "./service/config.json")

        self.__config = configuration.ServiceConfiguration(config_path, host_type)

        self.__service_instance_loader = service_loading.get_service_loader(self.__config.service())

        self.__plugins = list(self.__config.plugins())

        self.__schema: configuration.ServiceSchema = self.__config.schema()
        self.__info: configuration.ServiceInfo = self.__config.info()
        
        self.__service: SafeServiceWrapper or None = None

        self.__logger = logging.getLogger(self.__service_instance_loader.get_name())
        
    def get_info(self) -> configuration.ServiceInfo:
        return self.__info

    def get_load_parameter_specs(self) -> configuration.ParametersSchema:
        return self.__schema.parameters_load()
        
    def get_process_parameter_specs(self) -> configuration.ParametersSchema:
        return self.__schema.parameters_process()
        
    def get_input_dataset_specs(self) -> typing.Iterable[configuration.DatasetSchema]:
        return self.__schema.datasets_input()

    def get_output_dataset_specs(self) -> typing.Iterable[configuration.DatasetSchema]:
        return self.__schema.datasets_output()

    def get_parameter_real_path_value(self, name: str) -> os.PathLike:
        return self.__config.parameters().get_real_path_value(name)
    
    def __get_service_instance(self) -> SafeServiceWrapper:

        service = self.__service_instance_loader.get_instance()

        print("Got service: {}".format(service.__class__.__name__))

        service = SafeServiceWrapper(service)

        for plugin in self.__plugins:
            plugin_loader = service_loading.get_service_loader(plugin, [service])
            
            print(f"loading plugin {plugin_loader.get_name()}")

            service = SafeServiceWrapper(plugin_loader.get_instance())

        return service

    def is_loaded(self):
        return self.__service is not None

    def _build_load_context_source(self, include_environment_variables = True, override: dict = None):
        context_parts: typing.List[context_sources.ServiceContextSource] = []
        
        if override is not None:
            context_parts.append(context_sources.DictServiceContextSource(override))

        if include_environment_variables:
            context_parts.append(context_sources.EnvironmentVariableServiceContextSource("SERVICE_"))
        
        context_parts.append(_ServiceConfigurationServiceContextSource(self.__config))

        return context_sources.CoalescingServiceContextSource(context_parts)

    def build_load_context(self, ctx_source: context_sources.ServiceContextSource = None):
        if ctx_source is None:
            ctx_source = self._build_load_context_source()

        ctx = contexts.ServiceContext(ctx_source, self._get_load_logger())

        return ctx

    def _get_load_logger(self):
        return self.__logger.getChild("load")
        
    def _get_process_logger(self):
        return self.__logger.getChild("process")

    async def load(self, ctx_source: context_sources.ServiceContextSource = None):
        service = self.__get_service_instance()

        if service.has_load():
            ctx = self.build_load_context(ctx_source)

            print("service.load")
            await service.load(ctx)

        self.__service = service

    async def process(self, ctx_source: context_sources.ProcessContextSource):
        logger = self._get_process_logger()
        ctx = contexts.ProcessContext(ctx_source, logger)
        
        await self.__service.process(ctx)

    def dispose(self):
        if self.__service is None:
            return
        
        self.__service.dispose()
